#include "userControl.h"
#include <crypto/aes.h>


UserDatabase::UserDatabase() {
    authenticated = false;
}

UserDatabase &UserDatabase::getInstance() {
    static UserDatabase singleton;
    return singleton;
}

int UserDatabase::loadUserDatabase(const char *userDbFileLocation) {
    memset(userDatabase, 0, MAX_USER * NUM_USER_PARAM * MAX_PASSWORD_LENGTH);

    // save the database name, such that we can reload them later
    strcpy(databaseFsPath, userDbFileLocation);

    int dbsize;
    FILE *fptr;
    fptr = fopen(userDbFileLocation, "rb");

    if (!fptr) {
        // printf("cannot find database at location %s\nInitializing empty database\n", userDbFileLocation);
        initializeDefaultDatabase();
        return -1;
    }

    // read the first four bytes of the file to get the size of the database
    fscanf(fptr, "%d", &dbsize);

    // read the binary database back into database
    fread(encryptedUserDatabase, 1, dbsize, fptr);

    // close the file
    fclose(fptr);

    // use this to test bypassing encryption
    // memcpy((uint8_t*)userDatabase,
    // 		 (uint8_t*)encryptedUserDatabase,
    // 		 dbsize);

    //decrypt the database
    AES128_ECB_decrypt_buffer((uint8_t *) userDatabase,
                              (uint8_t *) encryptedUserDatabase,
                              dbsize,
                              (uint8_t *) mySuperSecretKey);

    return 0;
}

int UserDatabase::initializeDefaultDatabase() {
    strcpy(userDatabase[0][0], "admin");
    strcpy(userDatabase[0][1], "admin");
    saveDatabaseToDisk();
    return 0;
}

int UserDatabase::saveDatabaseToDisk() {
    // calculate the max size of the database
    int dbsize = MAX_USER * NUM_USER_PARAM * MAX_PASSWORD_LENGTH;

    // use this to test bypassing encryption
    // memcpy((uint8_t*)encryptedUserDatabase,
    // 						 (uint8_t*)userDatabase,
    // 						 dbsize);

    //encrypt the current database
    AES128_ECB_encrypt_buffer((uint8_t *) encryptedUserDatabase,
                              (uint8_t *) userDatabase,
                              dbsize,
                              (uint8_t *) mySuperSecretKey);

    // open the file
    FILE *fptr;
    fptr = fopen(databaseFsPath, "wb");

    fprintf(fptr, "%d", dbsize);
    fwrite(encryptedUserDatabase, 1, dbsize, fptr);
    fclose(fptr);
    return 0;
}

int UserDatabase::authenticateUser(const char *userName, const char *userPassword) {
    for (int i = 0; i < MAX_USER; i++) {
        if (strcmp(userDatabase[i][0], userName) == 0 &&
            strcmp(userDatabase[i][1], userPassword) == 0
                ) {
            authenticated = true;
            return 0;
        }
    }
    return 1;
}

int UserDatabase::addUser(const char *userToAdd, const char *newUserPassword) {
    if (!authenticated) return 1;


    for (int i = 0; i < MAX_USER; i++) {
        if (strcmp(userDatabase[i][0], userToAdd) == 0) {
            // printf("attempt to add user %s, that is already in the database\n",userToAdd);
            return 1;
        }

        if (strcmp(userDatabase[i][0], "") == 0) {
            strcpy(userDatabase[i][0], userToAdd);
            strcpy(userDatabase[i][1], newUserPassword);
            saveDatabaseToDisk();
            return 0;
        }

    }

    return 1;
}

int UserDatabase::removeUser(const char *userToRemove) {
    if (!authenticated) return 1;

    for (int i = 0; i < MAX_USER; i++) {
        if (strcmp(userDatabase[i][0], userToRemove) == 0) {
            memset(userDatabase[i][0], 0, MAX_PASSWORD_LENGTH);
            saveDatabaseToDisk();
            return 0;
        }

    }
    return 1;
}

bool UserDatabase::isAuthenticated() {
    return authenticated;
}