//
// Created by fangchenli on 7/22/19.
//

#ifndef TEE_LAB_KMEANS_H
#define TEE_LAB_KMEANS_H

#include <vector>
#include <iostream>
#include <sstream>
#include <algorithm>
#include <cmath>
#include <fstream>
#include <random>
#include <unordered_set>

using namespace std;

enum {
    MIN_K = 1,
    MIN_ARGS_KMEANS = 5,
    DEFAULT_K = 0,
    DEFAULT_ITER = 100
};

class Point {
public:

    int pointId, clusterId, dimension;
    vector<double> values;

    Point(int id, const string &line);

    int getId();

    int getClusterId();

    int getDimension();

    double getVal(int pos);

    void setCluster(int val);
};


class Cluster {
public:
    int clusterId;
    vector<double> centroid;
    vector<Point> points;

    Cluster(int clusterId, Point centroid);

    void addPoint(Point p);

    bool removePoint(int pointId);

    int getId();

    Point getPoint(int pos);

    int getSize();

    double getCentroidByPos(int pos);

    void setCentroidByPos(int pos, double val);
};


class KMeans {
private:
    int K, iters, dimension{}, num_points{};
    vector<Cluster> clusters;

    int getNearestClusterId(Point p);

public:

    KMeans();

    KMeans(int K, int iterations);

    void setK(int k);

    void setIters(int it);

    void run(vector<Point> &points);

    static int readData(const string &fileName, vector<Point> *points);

    void setClusters(vector<Cluster> vc);

    vector<int> prediction(const vector<Point>& points);

    string centroidsToString();
};

#endif //TEE_LAB_KMEANS_H
