#include "mlModule.h"
#include "crypto/cryptoUtil.h"

MlModule *MlModule::mlInstance = nullptr;

MlModule *MlModule::getInstance() {
    if (!mlInstance) {
        mlInstance = new MlModule;
    }
    return mlInstance;
}

// --------------------------------answer for kmeans--------------------------------------------

// save kmeans model
int MlModule::saveMlModel(const char *modelPath) {

    // write centroids to temp file
    std::ofstream out(temp);
    out << model_kmeans.centroidsToString();
    out.close();

    // encrypt the temp file
    CryptoUtil::secureEncryptFile(const_cast<char *>(modelPath), const_cast<char *>(temp));

    return SUCCESS;
}

// load kmeans model
int MlModule::loadSavedMlModel(const char *modelPath) {

    // decrypt the model to the temp file
    CryptoUtil::secureDecryptFile(const_cast<char *>(temp), const_cast<char *>(modelPath));

    ifstream infile;
    infile.open(temp, ios::in);
    if (infile.is_open()) {

        vector<Cluster> vc;
        string line;
        while (getline(infile, line)) {

            if (line.length() != 0) {
                istringstream iss(line);
                ostringstream oss;
                vector<string> tokens(istream_iterator<string>{iss}, istream_iterator<string>());

                // slice the vector to get the position of the centroid
                copy(tokens.begin() + 1, tokens.end() - 1, ostream_iterator<string>(oss, " "));
                oss << tokens.back();

                // reconstruct the cluster and push to the vector
                vc.emplace_back(Cluster(stoi(tokens[0]), Point(0, oss.str())));
            }
        }

        infile.close();

        // load model to the kmeans instance
        this->model_kmeans.setK(vc.size());
        this->model_kmeans.setIters(100);
        this->model_kmeans.setClusters(vc);
        return SUCCESS_OPEN_FILE;
    } else {
        return UNABLE_OPEN_FILE;
    }
}

// train kmeans model
int MlModule::trainMlModel(const char *trainingSamplePath) {

    // decrypt the sample to the temp file
    CryptoUtil::secureDecryptFile(const_cast<char *>(temp), const_cast<char *>(trainingSamplePath));

    // set value of K and number of iterations
    int K = 2, iters = 100;
    this->model_kmeans.setK(K);
    this->model_kmeans.setIters(iters);

    vector<Point> points;
    int dataFetchResult = KMeans::readData(temp, &points);
    if (dataFetchResult == SUCCESS_OPEN_FILE) {
        cout << "Data fetched success!" << endl;
    } else {
        cout << "Unable to open data file for k-means.";
        return dataFetchResult;
    }

    int points_size = points.size();
    if (points_size == 1) {
        cout << "There are only one data point. Please make check the data format." << endl;
        exit(1);
    }

    if (K > points_size) {
        cout << "K is greater than the number of data points." << endl;
        exit(1);
    }

    this->model_kmeans.run(points);

    return SUCCESS;
}

// predict with kmeans
int MlModule::makeInference(const char *samplePath) {

    // decrypt the sample to the temp file
    CryptoUtil::secureDecryptFile(const_cast<char *>(temp), const_cast<char *>(samplePath));    

    vector<Point> points;
    int dataFetchResult = KMeans::readData(temp, &points);
    if (dataFetchResult == SUCCESS_OPEN_FILE) {
        cout << "Data fetched success!" << endl;
    } else {
        cout << "Unable to open data file for k-means.";
        return dataFetchResult;
    }
    vector<int> predictionResult = model_kmeans.prediction(points);

    for (int result: predictionResult) {
        cout << result << endl;
    }

    return SUCCESS;
}
