// system parameters
// security headers
#include "accessControl/userControl.h"
#include "crypto/aes.h"
#include "crypto/cryptoUtil.h"
// ml headers
#include "ml/mlModule.h"

#include <iostream>
#include <cstring>

using std::cout;
using std::endl;

#ifdef _WIN32
#include <windows.h>

void mySleep(unsigned milliseconds)
{
    Sleep(milliseconds);
}
#else

#include <unistd.h>

void mySleep(unsigned milliseconds) {
    usleep(milliseconds * 1000); // takes microseconds
}

#endif

void unitTestAes();

void unitTestUserControl();

void unitTestCryptoUtil();

void encryptDataSet();

void unitTestKMeans();

void runUnitTest() {
    cout << "### Running Unit Tests" << endl;
    unitTestAes();
    unitTestUserControl();
    unitTestCryptoUtil();
    encryptDataSet();
    unitTestKMeans();
}

void setAesParam(unsigned char *key, unsigned char *iv, unsigned char *in, unsigned char *out, unsigned char *buffer) {
    unsigned char key0[] = {0x2b, 0x7e, 0x15, 0x16, 0x28, 0xae, 0xd2, 0xa6, 0xab, 0xf7, 0x15, 0x88, 0x09, 0xcf, 0x4f,
                            0x3c};
    unsigned char iv0[] = {0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e,
                           0x0f};

    unsigned char in0[] = {0x6b, 0xc1, 0xbe, 0xe2, 0x2e, 0x40, 0x9f, 0x96, 0xe9, 0x3d, 0x7e, 0x11, 0x73, 0x93, 0x17,
                           0x2a,
                           0xae, 0x2d, 0x8a, 0x57, 0x1e, 0x03, 0xac, 0x9c, 0x9e, 0xb7, 0x6f, 0xac, 0x45, 0xaf, 0x8e,
                           0x51,
                           0x30, 0xc8, 0x1c, 0x46, 0xa3, 0x5c, 0xe4, 0x11, 0xe5, 0xfb, 0xc1, 0x19, 0x1a, 0x0a, 0x52,
                           0xef,
                           0xf6, 0x9f, 0x24, 0x45, 0xdf, 0x4f, 0x9b, 0x17, 0xad, 0x2b, 0x41, 0x7b, 0xe6, 0x6c, 0x37,
                           0x10};

    unsigned char out0[] = {0x76, 0x49, 0xab, 0xac, 0x81, 0x19, 0xb2, 0x46, 0xce, 0xe9, 0x8e, 0x9b, 0x12, 0xe9, 0x19,
                            0x7d,
                            0x50, 0x86, 0xcb, 0x9b, 0x50, 0x72, 0x19, 0xee, 0x95, 0xdb, 0x11, 0x3a, 0x91, 0x76, 0x78,
                            0xb2,
                            0x73, 0xbe, 0xd6, 0xb8, 0xe3, 0xc1, 0x74, 0x3b, 0x71, 0x16, 0xe6, 0x9e, 0x22, 0x22, 0x95,
                            0x16,
                            0x3f, 0xf1, 0xca, 0xa1, 0x68, 0x1f, 0xac, 0x09, 0x12, 0x0e, 0xca, 0x30, 0x75, 0x86, 0xe1,
                            0xa7};
    memcpy(key, key0, 16);
    memcpy(iv, iv0, 16);
    memcpy(in, in0, 64);
    memcpy(out, out0, 64);
    memset(buffer, 0xFF, 64);
}

void unitTestAes() {
    unsigned char key[16];
    unsigned char iv[16];
    unsigned char in[64];
    unsigned char out[64];
    unsigned char buffer[64];
    const unsigned int buffer_length = 64;


    setAesParam(key, iv, in, out, buffer);
    AES128_CBC_encrypt_buffer(buffer, in, buffer_length, key, iv);
    if (memcmp(buffer, out, buffer_length) != 0)
        cout << "buffer is different, AES CBC encrypt implementation is broken" << endl;
    else
        cout << "OK - AES128 CBC encrypt" << endl;

    setAesParam(key, iv, in, out, buffer);
    AES128_CBC_decrypt_buffer(buffer, out, buffer_length, key, iv);
    if (memcmp(buffer, in, buffer_length) != 0)
        cout << "buffer is different, AES CBC decrypt implementation is broken" << endl;
    else
        cout << "OK - AES128 CBC decrypt" << endl;



    // ECB has a different ciphertext 
    uint8_t ecb_out[] = {0x3a, 0xd7, 0x7b, 0xb4, 0x0d, 0x7a, 0x36, 0x60, 0xa8, 0x9e, 0xca, 0xf3, 0x24, 0x66, 0xef,
                         0x97};

    setAesParam(key, iv, in, out, buffer);
    memcpy(out, ecb_out, 16);
    AES128_ECB_encrypt(in, key, buffer);
    if (memcmp(buffer, out, 16) != 0)
        cout << "buffer is different, AES ECB encrypt implementation is broken" << endl;
    else
        cout << "OK - AES128 ECB encrypt" << endl;

    setAesParam(key, iv, in, out, buffer);
    memcpy(out, ecb_out, 16);
    AES128_ECB_decrypt(out, key, buffer);
    if (memcmp(buffer, in, 16) != 0)
        cout << "buffer is different, AES ECB decrypt implementation is broken" << endl;
    else
        cout << "OK - AES128 ECB decrypt" << endl;

    setAesParam(key, iv, in, out, buffer);
    AES128_ECB_encrypt_buffer(buffer, in, buffer_length, key);
    if (memcmp(buffer, in, buffer_length) == 0) cout << "AES ECB buffer encrypt fail" << endl;
    AES128_ECB_decrypt_buffer(out, buffer, buffer_length, key);
    if (memcmp(in, out, buffer_length) != 0) cout << "AES ECB buffer decrypt fail" << endl;
    else cout << "OK - AES ECB buffer encrypt and decrypt" << endl;


}


void unitTestUserControl() {
    UserDatabase::getInstance().loadUserDatabase("testDb");
    if (UserDatabase::getInstance().addUser("root", "root") == 0)
        cout << "ERROR: able to add user before authentication" << endl;
    else if (UserDatabase::getInstance().authenticateUser("root", "root") == 0)
        cout << "ERROR: ilegal user authenticated" << endl;
    else if (UserDatabase::getInstance().removeUser("root") == 0)
        cout << "ERROR: able to remove user before authentication" << endl;
    else if (UserDatabase::getInstance().authenticateUser("admin", "admin") != 0)
        cout << "ERROR: unable to login via default credential" << endl;
    else if (UserDatabase::getInstance().addUser("admin", "admin") == 0)
        cout << "ERROR: able to add two user with the same id" << endl;
    else if (UserDatabase::getInstance().addUser("root", "root") != 0)
        cout << "ERROR: unable to add user" << endl;
    else if (UserDatabase::getInstance().authenticateUser("root", "root") != 0)
        cout << "ERROR: unable to authenticate a legal user" << endl;
    else if (UserDatabase::getInstance().removeUser("admin") != 0)
        cout << "ERROR: unable to remove a user user" << endl;
    else
        cout << "OK - user test" << endl;

}

void unitTestCryptoUtil() {
    unsigned char aesKey1[16];
    unsigned char aesKey2[16];

    CryptoUtil::getAesKey(aesKey1, 16);
    mySleep(100);
    CryptoUtil::getAesKey(aesKey2, 16);

    if (memcmp(aesKey1, aesKey2, 16) != 0)
        cout << "AES key gen is not random" << endl;
    else
        cout << "OK - AES key gen verification" << endl;

    char memBuffer[128];
    strcpy(memBuffer, "Mary had a little lamb");

    ofstream myfile;
    myfile.open("originalFile");
    myfile << memBuffer << "\n";
    myfile.close();

    CryptoUtil::secureEncryptFile((char *) "encryptedFile", (char *) "originalFile");
    CryptoUtil::secureDecryptFile((char *) "decryptedFile", (char *) "encryptedFile");

    string line;
    ifstream myInfile("decryptedFile");
    getline(myInfile, line);
    myInfile.close();

    if (strcmp(line.c_str(), memBuffer) != 0)
        cout << "file encryption and decryption fail, the decrypted content is\n"
             << line << endl;
    else
        cout << "OK - encrypt and decrypt file" << endl;
}

void unitTestKMeans() {

    cout << endl << "### Running KMeans Unit Test" << endl;

    const char *trainingDataPath = "../data/geo/points";
    const char *testFile = "../data/geo/test";
    const char *encryptedTrainingSamplesPath = "encryptedTrainingSamples";
    const char * encryptedTestData = "encryptedData";
    const char *outputFile = "clusters";

    CryptoUtil::secureEncryptFile((char *)encryptedTrainingSamplesPath, (char *)trainingDataPath);
    CryptoUtil::secureEncryptFile((char *)encryptedTestData, (char *)testFile);
    
    MlModule::getInstance()->trainMlModel(encryptedTrainingSamplesPath);

    cout << "OK - Train Model" << endl;

    MlModule::getInstance()->saveMlModel(outputFile);

    cout << "OK - Save Model" << endl;

    MlModule::getInstance()->loadSavedMlModel(outputFile);

    cout << "OK - Load Model" << endl;

    MlModule::getInstance()->makeInference(encryptedTestData);

    cout << "OK - Make Prediction" << endl;
}

/*
 * encrypt dataset
 */
void encryptDataSet() {

    // encrypt dataset for kmeans
    CryptoUtil::secureEncryptFile((char *)"../data_encrypted/geo/points", (char *)"../data/geo/points");
    CryptoUtil::secureEncryptFile((char *)"../data_encrypted/geo/test", (char *)"../data/geo/test");

    // encrypt dataset for libsvm
    CryptoUtil::secureEncryptFile((char *)"../data_encrypted/libsvm/a1a", (char *)"../data/libsvm/a1a");
    CryptoUtil::secureEncryptFile((char *)"../data_encrypted/libsvm/a1a.t", (char *)"../data/libsvm/a1a.t");

    // TODO assertion error
    // encrypt mnist for cnn
    CryptoUtil::secureEncryptFile((char *)"../data_encrypted/mnist/t10k-images.idx3-ubyte", (char *)"../data/mnist/t10k-images.idx3-ubyte");
    CryptoUtil::secureEncryptFile((char *)"../data_encrypted/mnist/t10k-labels.idx1-ubyte", (char *)"../data/mnist/t10k-labels.idx1-ubyte");
    CryptoUtil::secureEncryptFile((char *)"../data_encrypted/mnist/test.ppm",               (char *)"../data/mnist/test.ppm");
    CryptoUtil::secureEncryptFile((char *)"../data_encrypted/mnist/train-images.idx3-ubyte",(char *)"../data/mnist/train-images.idx3-ubyte");
    CryptoUtil::secureEncryptFile((char *)"../data_encrypted/mnist/train-labels.idx1-ubyte",(char *)"../data/mnist/train-labels.idx1-ubyte");
}